@@ -23,13 +23,11 @@ Annotate::Annotate(Git* parent, QObject* guiObj) : QObject(parent) {
 
 	processingTime.start();
 
-	connect(&patchProc, SIGNAL(readyReadStdout()),
-		    this, SLOT(patchProcReadFromStdout()));
-	connect(&patchProc, SIGNAL(processExited()),
-		    this, SLOT(patchProcExited()));
+	connect(&patchProc, SIGNAL(readyReadStdout()), this, SLOT(patchProcReadFromStdout()));
+	connect(&patchProc, SIGNAL(processExited()), this, SLOT(patchProcExited()));
 }
 
-const FileAnnotation * Annotate::lookupAnnotation(SCRef sha, SCRef fn) {
+const FileAnnotation* Annotate::lookupAnnotation(SCRef sha, SCRef fn) {
 
 	if (!valid || fileName != fn)
 		return NULL;
@@ -54,7 +52,7 @@ const FileAnnotation * Annotate::lookupAnnotation(SCRef sha, SCRef fn) {
 bool Annotate::startPatchProc(SCRef buf, SCRef fileName) {
 
 	const QString cmd("git-diff-tree -r -m --patch-with-raw "
-			"--no-commit-id --stdin -- " + fileName);
+	                  "--no-commit-id --stdin -- " + fileName);
 	patchProc.setArguments(QStringList::split(' ', cmd));
 	patchProc.setWorkingDirectory(git->workDir);
 	patchProcBuf = "";
@@ -83,8 +81,10 @@ bool Annotate::stop() {
 
 void Annotate::slotProgressInfo() {
 
-	if (!cancelingAnnotate && !isError)
-		QApplication::postEvent(gui, new AnnotateProgressEvent(annFilesNum));
+	if (!cancelingAnnotate && !isError) {
+		const QString n(QString::number(annFilesNum));
+		QApplication::postEvent(gui, new AnnotateProgressEvent(n));
+	}
 }
 
 bool Annotate::start(SCRef _fileName, const StrVect& _histRevOrder) {
@@ -108,7 +108,6 @@ bool Annotate::start(SCRef _fileName, const StrVect& _histRevOrder) {
 		ah.insert(*it, FileAnnotation(annId--));
 	while (++it != histRevOrder.constEnd());
 
-
 	// annotation is splitted in two parts.
 	// first we get the list of sha's to feed git-diff-tree.
 	// This is very fast.
@@ -158,23 +157,25 @@ void Annotate::computeDiffs() {
 	cancelingAnnotate = annotateRunning = false;
 	git->annotateExited(this);
 
-// 	StrVect::const_iterator it(histRevOrder.constBegin()); // DEBUG
+// 	StrVect::const_iterator it(histRevOrder.constBegin());
 // 	do {
-// 		dbg(*it);dbg(ah[*it].fileSha);
+// 		dbg(*it); dbg(ah[*it].fileSha);
 // 	} while (++it != histRevOrder.constEnd());
 }
 
 void Annotate::annotateFileHistory(SCRef fileName, bool buildPatchScript) {
 
-	// we cannot rely on pure recursive doAnnotate() because
-	// we can have more then one branch in which fileName appears
-	StrVect::const_iterator it(histRevOrder.constBegin());
-	do
+	// sweep from the oldest to newest so that parent
+	// annotations are calculated before childern
+	StrVect::const_iterator it(histRevOrder.constEnd());
+	do {
+		--it;
 		doAnnotate(fileName, *it, buildPatchScript);
-	while (++it != histRevOrder.constEnd());
+	} while (it != histRevOrder.constBegin() && !isError);
 }
 
 void Annotate::doAnnotate(SCRef fileName, SCRef sha, bool buildPatchScript) {
+// all the parents annotations must be valid here
 
 	FileAnnotation* fa = getFileAnnotation(sha);
 	if (fa == NULL || fa->isValid || isError)
@@ -192,22 +193,16 @@ void Annotate::doAnnotate(SCRef fileName, SCRef sha, bool buildPatchScript) {
 		fa->isValid = true;
 		return;
 	}
-	// be sure we have all the parents annotations ready
-	const QStringList parents(r->parents());
-	QStringList::const_iterator it(parents.constBegin());
-	while (it != parents.constEnd()) {
-		FileAnnotation* pa = getFileAnnotation(*it);
-		if (!pa->isValid) {
-			doAnnotate(fileName, *it, buildPatchScript); // recursive call
-			if (cancelingAnnotate || isError)
-				return;
-		}
-		++it;
-	}
 	// now create a new annotation from first parent diffs
+	const QStringList parents(r->parents());
 	const QString& parSha = parents.first();
 	FileAnnotation* pa = getFileAnnotation(parSha);
 
+	if (!pa->isValid) {
+		dbp("ASSERT in doAnnotate: annotation for %1 not valid", parSha);
+		isError = true;
+		return;
+	}
 	if (buildPatchScript) // just prepare patch script
 		updatePatchScript(sha, parSha);
 	else {
@@ -216,7 +211,7 @@ void Annotate::doAnnotate(SCRef fileName, SCRef sha, bool buildPatchScript) {
 		setAnnotation(diff, author, pa->lines, fa->lines);
 	}
 	// then add other parents diff if any
-	it = parents.constBegin();
+	QStringList::const_iterator it(parents.constBegin());
 	++it;
 	while (it != parents.constEnd()) {
 
@@ -234,8 +229,7 @@ void Annotate::doAnnotate(SCRef fileName, SCRef sha, bool buildPatchScript) {
 		// the two annotations must be of the same length
 		if (fa->lines.count() != tmpAnn.count()) {
 			qDebug("ASSERT: merging annotations of different length\n"
-					" merging %s in %s", (*it).latin1(), sha.latin1());
-
+			       " merging %s in %s", (*it).latin1(), sha.latin1());
 			isError = true;
 			return;
 		}
@@ -299,15 +293,13 @@ void Annotate::unify(SList dst, SCList src) {
 
 	QStringList::Iterator itd(dst.begin());
 	QStringList::const_iterator its(src.constBegin());
-	for ( ; itd != dst.end(); ++itd, ++its) {
+	for ( ; itd != dst.end(); ++itd, ++its)
 		if (*itd == "Merge")
 			*itd = *its;
-	}
 }
 
 void Annotate::setAnnotation(SCRef diff, SCRef author, SCList prevAnn,
-					    QStringList& newAnn, int ofs) {
-
+                             QStringList& newAnn, int ofs) {
 	newAnn = prevAnn;
 	QStringList::iterator cur(newAnn.begin());
 	QString line;
@@ -353,21 +345,22 @@ void Annotate::setAnnotation(SCRef diff, SCRef author, SCList prevAnn,
 				if (cur != newAnn.end())
 					cur = newAnn.remove(cur);
 				else {
-					dbp("ASSERT processDiff: remove end of file, diff is %1", diff);
+					dbp("ASSERT processDiff: remove end of "
+					    "file, diff is %1", diff);
 					isError = true;
 					return;
 				}
 			} else {
 				dbp("ASSERT processDiff: remove line from "
-					"empty annotation, diff is %1", diff);
+				    "empty annotation, diff is %1", diff);
 				isError = true;
 				return;
 			}
 			break;
 		case '\\':
-			// diff(1) produces a "\ No newline at end of file", but
-			// the message is locale dependent, so just test the space after '\'
-			if (line[1].latin1() == ' ')
+			// diff(1) produces a "\ No newline at end of file", but the
+			// message is locale dependent, so just test the space after '\'
+			if (line[1] == ' ')
 				break;
 		default:
 			++cur;
@@ -378,8 +371,8 @@ void Annotate::setAnnotation(SCRef diff, SCRef author, SCList prevAnn,
 
 void Annotate::updatePatchScript(SCRef sha, SCRef par) {
 
-	if (sha == QGit::ZERO_SHA) // diff-tree --stdin doesn't work with working dir
-		return; // patch will be directly fetched when needed
+	if (sha == QGit::ZERO_SHA) // diff-tree --stdin doesn't work with working
+		return;            // dir patch will be directly fetched when needed
 
 	const QString runCmd(sha + " " + par);
 	patchScript.append(runCmd).append('\n');
@@ -388,14 +381,14 @@ void Annotate::updatePatchScript(SCRef sha, SCRef par) {
 const QString Annotate::getNextPatch(QString& patchFile, SCRef fileName, SCRef sha) {
 
 	if (sha == QGit::ZERO_SHA) {
-		// diff-tree --stdin doesn't work with working dir
-		// so get it directly. We don't need file sha info because is ZERO_SHA
+		// diff-tree --stdin doesn't work with working dir so get it
+		// directly. We don't need file sha info because is ZERO_SHA
 		QString runCmd("git-diff-index -r -m -p HEAD -- " + fileName), runOutput;
 		git->run(runCmd, &runOutput);
 
 		// restore removed head line of next patch
 		const QString nextHeader("\n:100644 100644 " + QGit::ZERO_SHA + ' '
-				+ nextFileSha + " M\t" + fileName + '\n');
+		                         + nextFileSha + " M\t" + fileName + '\n');
 		runOutput.append(nextHeader);
 		patchFile.prepend(runOutput);
 		nextFileSha = QGit::ZERO_SHA;
