@@ -17,8 +17,7 @@ class Git;
 class CommitImpl : public CommitBase {
 Q_OBJECT
 public:
-	CommitImpl(Git* git, bool* changesCommitted, QWidget *parent = 0,
-	           const char *name = 0, WFlags fl = 0);
+	CommitImpl(Git* git, bool* changesCommitted);
 	~CommitImpl();
 
 public slots:
