@@ -17,11 +17,12 @@ class ListView;
 class FileContent;
 class MainImpl;
 class Git;
+class TabFile;
 
 class FileView: public Domain {
 Q_OBJECT
 public:
-	FileView(MainImpl* m, Git* git, QWidget* t);
+	FileView(MainImpl* m, Git* git, TabFile* t);
 	virtual ~FileView();
 	void append(SCRef data);
 	void clear(bool complete = true);
@@ -57,7 +58,7 @@ private:
 
 	EM_DECLARE(exDtorCalled);
 
-	QWidget* tabPage;
+	TabFile* fileTab;
 	ListView* histListView;
 	FileContent* textEditFile;
 };
