@@ -6,8 +6,10 @@
 	Copyright: See COPYING file that comes with this distribution
 
 */
-#include <unistd.h> // usleep()
-#include <stdlib.h> // used by getenv(), setenv()
+#include <unistd.h>    // usleep()
+#include <stdlib.h>    // used by getenv(), setenv()
+#include <sys/types.h> // used by chmod()
+#include <sys/stat.h>  // used by chmod()
 #include <qapplication.h>
 #include <qdatetime.h>
 #include <qtextcodec.h>
@@ -18,6 +20,7 @@
 #include <qeventloop.h>
 #include <qprocess.h>
 #include <qtextcodec.h>
+#include <qstylesheet.h>
 #include "mainimpl.h"
 #include "annotate.h"
 #include "cache.h"
@@ -176,7 +179,7 @@ bool Git::isUnapplied(SCRef sha) { return unAppliedSHA.contains(sha); }
 Annotate* Git::startAnnotate(QObject* guiObj) { // non blocking
 
 	Annotate* ann = new Annotate(this, guiObj);
-	ann->start(histFileName, histRevOrder); // non  blocking call
+	ann->start(histFileName, histRevOrder); // non blocking call
 	return ann;
 }
 
@@ -222,10 +225,10 @@ const Rev* Git::revLookup(SCRef sha, bool isHistoryView) {
 	return r[sha];
 }
 
-bool Git::run(SCRef runCmd, QString* runOutput, SCRef buf) {
+bool Git::run(SCRef runCmd, QString* runOutput, QObject* receiver, SCRef buf) {
 
 	MyProcess p(par, this, workDir, errorReportingEnabled);
-	return p.runSync(runCmd, runOutput, buf);
+	return p.runSync(runCmd, runOutput, receiver, buf);
 }
 
 MyProcess* Git::runAsync(SCRef runCmd, QObject* receiver, SCRef buf) {
@@ -238,6 +241,25 @@ MyProcess* Git::runAsync(SCRef runCmd, QObject* receiver, SCRef buf) {
 	return p; // auto-deleted when done
 }
 
+MyProcess* Git::runAsScript(SCRef runCmd, QObject* receiver, SCRef buf) {
+
+	const QString scriptFile(workDir + "/qgit_script.sh");
+	if (!writeToFile(scriptFile, runCmd))
+		return NULL;
+
+	chmod(scriptFile, 0755);
+	MyProcess* p = runAsync(scriptFile, receiver, buf);
+	if (p)
+		connect(p, SIGNAL(eof()), this, SLOT(on_runAsScript_eof()));
+	return p;
+}
+
+void Git::on_runAsScript_eof() {
+
+	QDir dir(workDir);
+	dir.remove("qgit_script.sh");
+}
+
 void Git::cancelProcess(MyProcess* p) {
 
 	if (p)
@@ -262,14 +284,13 @@ int Git::findFileIndex(const RevFile& rf, SCRef name) {
 
 const QString Git::getLaneParent(SCRef fromSHA, uint laneNum) {
 
-	QValueVector<QString>::const_iterator it;
-	it = qFind(revOrder.constBegin(), revOrder.constEnd(), fromSHA);
-	if (it == revOrder.constEnd())
+	const Rev* rs = revLookup(fromSHA);
+	if (!rs)
 		return QString();
 
-	while (it != revOrder.constBegin()) {
-		--it;
-		const Rev* r = revLookup(*it);
+	for (int idx = rs->orderIdx - 1; idx >= 0; idx--) {
+
+		const Rev* r = revLookup(revOrder[idx]);
 		if (laneNum >= r->lanes.count())
 			return QString();
 
@@ -289,16 +310,18 @@ const QString Git::getLaneParent(SCRef fromSHA, uint laneNum) {
 const QStringList Git::getChilds(SCRef parent) {
 
 	QStringList childs;
-	QDictIterator<Rev> it(revs);
-	for ( ; it.current(); ++it) { // traverse the whole map!
-		if (it.current()->parents().findIndex(parent) != -1)
-			childs.append(it.currentKey());
-	}
-	// reorder childs by id
+	const Rev* r = revLookup(parent);
+	if (!r)
+		return childs;
+
+	for (uint i = 0; i < r->childs.count(); i++)
+		childs.append(revOrder[r->childs[i]]);
+
+	// reorder childs by loading order
 	QStringList::iterator itC(childs.begin());
 	for ( ; itC != childs.end(); ++itC) {
 		const Rev* r = revLookup(*itC);
-		(*itC).prepend(QString("%1 ").arg(r->id, 5));
+		(*itC).prepend(QString("%1 ").arg(r->orderIdx, 5));
 	}
 	childs.sort();
 	for (itC = childs.begin(); itC != childs.end(); ++itC)
@@ -398,7 +421,7 @@ MyProcess* Git::getFile(SCRef file, SCRef revSha, QObject* receiver, QString* re
 }
 
 void Git::getTree(SCRef treeSha, SList names, SList shas,
-			   SList types, bool isWorkingDir, SCRef treePath) {
+                  SList types, bool isWorkingDir, SCRef treePath) {
 
 	QStringList newFiles, delFiles, dummy, tmp;
 	if (isWorkingDir) { // retrieve unknown and deleted files under treePath
@@ -512,6 +535,55 @@ bool Git::isSameFiles(SCRef tree1Sha, SCRef tree2Sha) {
 	return true;
 }
 
+const QStringList Git::getDescBranches(SCRef sha) {
+
+	QStringList tl;
+	const Rev* r = revLookup(sha);
+	if (!r || (r->descBrnMaster == -1))
+		return tl;
+
+	const QValueVector<int>& nr = revLookup(revOrder[r->descBrnMaster])->descBranches;
+
+	for (uint i = 0; i < nr.count(); i++) {
+
+		SCRef sha = revOrder[nr[i]];
+		int idx = headsSHA.findIndex(sha);
+		if (idx != -1) {
+			SCRef branchName = heads[idx];
+			const QString t = branchName + " (" + sha + ")";
+			tl.append(t);
+		}
+	}
+	return tl;
+}
+
+const QStringList Git::getNearTags(bool down, SCRef sha) {
+
+	QStringList tl;
+	const Rev* r = revLookup(sha);
+	if (!r)
+		return tl;
+
+	int nearRefsMaster = (down) ? r->descRefsMaster : r->ancRefsMaster;
+	if (nearRefsMaster == -1)
+		return tl;
+
+	const QValueVector<int>& nr = down ? revLookup(revOrder[nearRefsMaster])->descRefs :
+	                                     revLookup(revOrder[nearRefsMaster])->ancRefs;
+
+	for (uint i = 0; i < nr.count(); i++) {
+
+		SCRef sha = revOrder[nr[i]];
+		int idx = tagsSHA.findIndex(sha);
+		if (idx != -1) {
+			SCRef tagName = tags[idx];
+			const QString t = tagName + " (" + sha + ")";
+			tl.append(t);
+		}
+	}
+	return tl;
+}
+
 const QString Git::getDefCommitMsg() {
 
 	SCRef sha((appliedSHA.empty()) ? ZERO_SHA : appliedSHA.last());
@@ -556,18 +628,40 @@ const QString Git::getDesc(SCRef sha, QRegExp& shortLogRE, QRegExp& longLogRE) {
 		dbp("ASSERT: getDesc sha <%1> not found", sha);
 		return QString();
 	}
+
+	// set Latin-1 to avoid using QString::fromLatin1() everywhere
+	QTextCodec* tc = QTextCodec::codecForCStrings();
+	QTextCodec::setCodecForCStrings(0);
+
+	QString text;
 	if (c->isDiffCache)
-		return c->longLog();
+		text = c->longLog();
+	else {
+		text = QString("Author: " + c->author() + "\nDate:   ");
+		text.append(getLocalDate(c->authorDate()));
+		if (!c->isUnApplied) {
+			text.append("\nParent: ").append(c->parents().join("\nParent: "));
 
-	QString text("Author: " + c->author() + "\nDate:   ");
-	text.append(getLocalDate(c->authorDate()));
-	text.append("\nParent: ").append(c->parents().join("\nParent: "));
-	text.append("\n\n    " + colorMatch(c->shortLog(), shortLogRE) +
-			'\n' + colorMatch(c->longLog(), longLogRE));
+			QStringList ch = getChilds(sha);
+			if (!ch.isEmpty())
+				text.append("\nChild: ").append(ch.join("\nChild: "));
 
-	// replace '<' and '>' to avoid fooling rtf engine
-	text.replace('<', QString::fromLatin1("&lt;"));
-	text.replace('>', QString::fromLatin1("&gt;"));
+			QStringList tg = getDescBranches(sha);
+			if (!tg.empty())
+				text.append("\nBranch: ").append(tg.join("\nBranch: "));
+
+			tg = getNearTags(false, sha);
+			if (!tg.isEmpty())
+				text.append("\nFollows: ").append(tg.join(", "));
+
+			tg = getNearTags(true, sha);
+			if (!tg.isEmpty())
+				text.append("\nPrecedes: ").append(tg.join(", "));
+		}
+		text.append("\n\n    " + colorMatch(c->shortLog(), shortLogRE) +
+		            '\n' + colorMatch(c->longLog(), longLogRE));
+	}
+	text = QStyleSheet::convertFromPlainText(text); // this puppy needs Latin-1
 
 	// hightlight SHA's
 	int pos = 0;
@@ -583,29 +677,15 @@ const QString Git::getDesc(SCRef sha, QRegExp& shortLogRE, QRegExp& longLogRE) {
 		if (slog.length() > 60)
 			slog = slog.left(57).stripWhiteSpace().append("...");
 
-		slog.replace('<', QString::fromLatin1("&lt;"));
-		slog.replace('>', QString::fromLatin1("&gt;"));
-		const QString link(QString::fromLatin1("<a href=\"") + sha +
-			QString::fromLatin1("\">") + slog + QString::fromLatin1("</a>"));
-
+		slog = QStyleSheet::escape(slog);
+		const QString link("<a href=\"" + sha + "\">" + slog + "</a>");
 		text.replace(pos, sha.length(), link);
 		pos += link.length();
 	}
+	text.replace("$_1", "<"); // see colorMatch()
+	text.replace("$_2", ">");
 
-	// convert spaces and tabs to &nbsp; to preserve multiple spaces
-	pos = 0;
-	QRegExp longSpaces(" ([ ]+)");
-	while ((pos = longSpaces.search(text)) != -1) {
-		QString cap(longSpaces.cap(1));
-		cap.replace(QRegExp(" "), QString::fromLatin1("&nbsp;"));
-		text.replace(pos + 1, longSpaces.matchedLength() - 1, cap);
-	}
-	text.replace(QRegExp("\t"), QString::fromLatin1("&nbsp;&nbsp;&nbsp;"));
-	text.replace(QRegExp("\n"), QString::fromLatin1("<br>\n"));
-	text.replace(QString::fromLatin1("$_1"), QString::fromLatin1("<"));
-	text.replace(QString::fromLatin1("$_2"), QString::fromLatin1(">"));
-	text.prepend(QString::fromLatin1("<qt><p>"));
-	text.append(QString::fromLatin1("</p></qt>"));
+	QTextCodec::setCodecForCStrings(tc); // restore codec
 	return text;
 }
 
@@ -716,7 +796,7 @@ bool Git::getPatchFilter(SCRef exp, bool isRegExp, QMap<QString, bool>& shaMap)
 		runCmd.append("--pickaxe-regex ");
 
 	runCmd.append(QUOTE_CHAR + "-S" + exp + QUOTE_CHAR);
-	if (!run(runCmd, &runOutput, buf))
+	if (!run(runCmd, &runOutput, NULL, buf))
 		return false;
 
 	QStringList sl(QStringList::split('\n', runOutput));
@@ -726,27 +806,36 @@ bool Git::getPatchFilter(SCRef exp, bool isRegExp, QMap<QString, bool>& shaMap)
 	return true;
 }
 
-bool Git::applyPatchFile(SCRef patchPath, bool dropping, bool fold) {
+bool Git::resetCommits(int parentDepth) {
 
-	SCRef quotedPath(QUOTE_CHAR + patchPath + QUOTE_CHAR);
+	QString runCmd("git-reset --soft HEAD~");
+	runCmd.append(QString::number(parentDepth));
+	return run(runCmd);
+}
 
-	if (isStGIT) {
+bool Git::applyPatchFile(SCRef patchPath, bool commit, bool fold, bool dropping) {
+
+	const QString quotedPath(QUOTE_CHAR + patchPath + QUOTE_CHAR);
+
+	if (commit && isStGIT) {
 		if (fold)
 			return run("stg fold " + quotedPath);
 
 		return run("stg import --mail " + quotedPath);
 	}
 	QString runCmd("git-am -k -u --3way ");
-	if (testFlag(SIGN_F) && !dropping)
+
+	if (testFlag(SIGN_PATCH_F) && !dropping)
 		runCmd.append("--signoff ");
-	return run(runCmd  + quotedPath);
+
+	return run(runCmd + quotedPath);
 }
 
 bool Git::formatPatch(SCList shaList, SCRef dirPath, SCRef remoteDir) {
 
 	bool remote = !remoteDir.isEmpty();
 	QSettings settings;
-	const QString FPArgs = settings.readEntry(APP_KEY + FPATCH_ARGS_KEY, FPATCH_ARGS_DEF);
+	const QString FPArgs = settings.readEntry(APP_KEY + FPATCH_ARGS_KEY, "");
 
 	QString runCmd("git-format-patch");
 	if (testFlag(NUMBERS_F) && !remote)
@@ -757,21 +846,19 @@ bool Git::formatPatch(SCList shaList, SCRef dirPath, SCRef remoteDir) {
 	if (!FPArgs.isEmpty())
 		runCmd.append(" " + FPArgs);
 
-// 	runCmd.append(" --start-number="); keep git 1.3.0 compatibility
+	runCmd.append(" --start-number=");
 
 	const QString tmp(workDir);
 	if (remote)
-		workDir = remoteDir;
+		workDir = remoteDir; // run() uses workDir value
 
-// 	int n = 1; keep git 1.3.0 compatibility
+	int n = 1;
 	bool ret = false;
 	loopList(it, shaList) { // shaList is ordered by newest to oldest
-/*		const QString cmd(runCmd + QString::number(n) + " " +
+		const QString cmd(runCmd + QString::number(n) + " " +
 				*it + QString::fromLatin1("^..") + *it);
-		n++;*/
-		const QString cmd(runCmd + " " +
-				*it + QString::fromLatin1("^..") + *it);
-		ret = run(cmd); // run() uses workDir value
+		n++;
+		ret = run(cmd);
 		if (!ret)
 			break;
 	}
@@ -783,6 +870,7 @@ bool Git::updateIndex(SCList selFiles) {
 
 	if (selFiles.empty())
 		return true;
+
 	QString runCmd("git-update-index --add --remove --replace -- ");
 	runCmd.append(selFiles.join(" "));
 	return run(runCmd);
@@ -799,7 +887,7 @@ bool Git::commitFiles(SCList selFiles, SCRef msg) {
 
 	// add user selectable commit options
 	QSettings settings;
-	const QString CMArgs = settings.readEntry(APP_KEY + CMT_ARGS_KEY, CMT_ARGS_DEF);
+	const QString CMArgs = settings.readEntry(APP_KEY + CMT_ARGS_KEY, "");
 
 	QString cmtOptions;
 	if (!CMArgs.isEmpty())
@@ -847,7 +935,7 @@ bool Git::commitFiles(SCList selFiles, SCRef msg) {
 		if (!updateIndex(notSelCachedFiles))
 			return false;
 	QDir dir(workDir);
-  	dir.remove(msgFile);
+	dir.remove(msgFile);
 	return true;
 }
 
@@ -970,7 +1058,7 @@ exit:
 	restoreFiles(notSelFiles, workDir, extNS);
 	updateIndex(notSelCachedFiles);
 	QDir dir(workDir);
-  	dir.remove(patchFile);
+	dir.remove(patchFile);
 	loopList(it, selFiles)
 		dir.remove(*it + extS); // remove temporary backup rollback files
 	return retval;
@@ -1007,6 +1095,7 @@ bool Git::stgPop(SCRef sha) {
 
 	if (sha == appliedSHA.last()) // top of the stack
 		return run("stg pop");
+
 	if (sha == appliedSHA.first())
 		return run("stg pop --all");
 
