@@ -67,7 +67,7 @@ const QString Git::getBaseDir(bool* changed, SCRef wd, bool* ok, QString* gd) {
 		return wd;
 	}
 	// 'git-rev-parse --git-dir' output could be a dir relative
-	//  to working dir (as ex .git) or an absolute path.
+	// to working dir (as ex .git) or an absolute path.
  	QDir d(runOutput.startsWith("/") ? runOutput : wd + "/" + runOutput);
 	*changed = (d.absPath() != gitDir);
 	if (gd)
@@ -179,7 +179,7 @@ void Git::dirWalker(SCRef dirPath, SList files, SList filesSHA, SCRef nameFilter
 			filesSHA.append(sha);
 		}
 		++it;
-    }
+	}
 }
 
 bool Git::lookUpPatchesSHA(SCRef patchName, SList files, SList filesSHA, QString& sha) {
@@ -188,7 +188,7 @@ bool Git::lookUpPatchesSHA(SCRef patchName, SList files, SList filesSHA, QString
 	QStringList fl(files.grep("/" + patchName + "/top"));
 	if (fl.count() != 1) {
 		qDebug("ASSERT: found %i patches instead of 1 in %s",
-			  fl.count(), patchName.latin1());
+		       fl.count(), patchName.latin1());
 		return false;
 	}
 	int pos = files.findIndex(fl.first());
@@ -247,7 +247,7 @@ bool Git::getStGITPatches() {
 	return true;
 }
 
-const Rev* Git::fakeWorkDirRev(SCRef parent, SCRef log, SCRef longLog) {
+const Rev* Git::fakeWorkDirRev(SCRef parent, SCRef log, SCRef longLog, int idx) {
 
 	QString date(QString::number(QDateTime::currentDateTime().toTime_t()) + " +0200");
 	QString data(ZERO_SHA + ' ' + parent + "\ntree ");
@@ -258,10 +258,9 @@ const Rev* Git::fakeWorkDirRev(SCRef parent, SCRef log, SCRef longLog) {
 	data.append("\n\n    " + log + '\n');
 	data.append(longLog);
 
-	Rev* c = new Rev(data);
+	Rev* c = new Rev(data, idx);
 
 	c->isDiffCache = true;
-	c->id = FIRST_BR_ID; // same branch of HEAD
 	c->lanes.append(EMPTY);
 	return c;
 }
@@ -319,13 +318,14 @@ void Git::getDiffIndex() {
 		return;
 
 	const QString parent(runOutput.section('\n', 0, 0));
-	revs.insert(ZERO_SHA, fakeWorkDirRev(parent, log, status));
+	revs.insert(ZERO_SHA, fakeWorkDirRev(parent, log, status, revs.count()));
 
 	// check for files already updated in cache, we will
 	// save this information as a prefix in status field
 	RevFile cachedFiles;
 	if (!run("git-diff-index --cached HEAD", &runOutput))
 		return;
+
 	parseDiffFormat(cachedFiles, runOutput);
 
 	for (uint i = 0; i < (*itf).status.count(); i++) {
@@ -524,14 +524,14 @@ bool Git::init(SCRef wd, bool askForRange, QStringList* filterList, bool* quit)
 				if (startUnappliedList()) {
 					EM_BEFORE_PROCESS_EVENTS;
 					while (loadingUnAppliedPatches) {
-						// WARNING this keeps us in setRepository()
+						// WARNING we are in setRepository()
 						usleep(20000);
 						qApp->processEvents();
 					}
 					EM_AFTER_PROCESS_EVENTS;
 					lns.clear(); // again to reset lanes
 				} else
-					qDebug("ERROR: unable to start unapplied list loading");
+					qDebug("ERROR: unable to load unapplied list");
 			}
 		}
 		if (testFlag(DIFF_INDEX_F) && !filteredLoading) {
@@ -554,7 +554,7 @@ bool Git::init(SCRef wd, bool askForRange, QStringList* filterList, bool* quit)
 			return false;
 		}
 		const QString info("Exception \'" + EM_DESC(i) + "\' "
-				"not handled in init...re-throw");
+		                   "not handled in init...re-throw");
 		dbp("%1", info);
 		throw i;
 	}
@@ -571,7 +571,7 @@ void Git::on_newHistDataReady() {
 }
 
 void Git::on_loaded(ulong byteSize, const QTime& loadTime,
-				bool normalExit, SCRef cmd, SCRef errorDesc) {
+                    bool normalExit, SCRef cmd, SCRef errorDesc) {
 
 	if (!errorDesc.isEmpty()) {
 		MainExecErrorEvent* e = new MainExecErrorEvent(cmd, errorDesc);
@@ -584,8 +584,8 @@ void Git::on_loaded(ulong byteSize, const QTime& loadTime,
 		if (!loadingUnAppliedPatches) {
 
 			uint kb = byteSize / 1024;
-			QString tmp(QString("Loaded %1 revisions (%2 KB),  time elapsed: %3 ms")
-					.arg(revs.count()).arg(kb).arg(loadTime.elapsed()));
+			QString tmp(QString("Loaded %1 revisions (%2 KB), time elapsed: %3 ms")
+			                    .arg(revs.count()).arg(kb).arg(loadTime.elapsed()));
 
 			emit loadCompleted(tmp);
 
@@ -613,10 +613,11 @@ void Git::loadFileNames() {
 		const QString runCmd("git-diff-tree -r -c --stdin");
 		runAsync(runCmd, this, diffTreeBuf);
 	}
+	indexTree();
 }
 
-void Git::on_histLoaded(ulong, const QTime&,
-				bool normalExit, SCRef cmd, SCRef errorDesc) {
+void Git::on_histLoaded(ulong, const QTime&, bool normalExit,
+                        SCRef cmd, SCRef errorDesc) {
 
 	if (!errorDesc.isEmpty()) {
 		MainExecErrorEvent* e = new MainExecErrorEvent(cmd, errorDesc);
@@ -658,7 +659,9 @@ void Git::addChunk(bool addToHistory, SCRef buffer) {
 			return;
 		dbp("ASSERT: addChunk sha <%1> already received", sha);
 	}
-	r.insert(sha, new Rev(buffer)); // only here we create a new rev
+
+	r.insert(sha, new Rev(buffer, r.count())); // only here we create a new rev
+
 	Rev* c = const_cast<Rev*>(revLookup(sha, addToHistory));
 	updateRefs(*c, addToHistory);
 
@@ -702,16 +705,11 @@ void Git::copyDiffIndex(RevMap& histRevs, SCRef histFileName) {
 	// insert a custom ZERO_SHA rev with proper parent
 	QDictIterator<Rev> it(histRevs); // only one item
 	QString parent(it.currentKey());
-	histRevs.insert(ZERO_SHA, fakeWorkDirRev(parent, "Working dir changes", "dummy"));
+	histRevs.insert(ZERO_SHA, fakeWorkDirRev(parent, "Working dir changes",
+	                                         "dummy", histRevs.count()));
 	histRevOrder.append(ZERO_SHA);
 }
 
-int Git::getId() {
-
-	static int id = FIRST_BR_ID;
-	return --id; // ordered from oldest to newest
-}
-
 void Git::setLane(SCRef sha, bool isHistory) {
 
 	Lanes& l = (isHistory) ? histLns : lns;
@@ -734,7 +732,6 @@ void Git::setLane(SCRef sha, bool isHistory) {
 void Git::updateLanes(Rev& c, Lanes& lns) {
 
 	SCRef sha(c.sha());
-	static int prevId = FIRST_BR_ID;
 	if (lns.isEmpty())
 		lns.init(sha);
 
@@ -744,13 +741,11 @@ void Git::updateLanes(Rev& c, Lanes& lns) {
 	bool isInitial = (c.parentsCount() == 0);
 	bool isApplied = (c.isApplied);
 
-	if (isDiscontinuity) {
+	if (isDiscontinuity)
 		lns.changeActiveLane(sha); // uses previous isBoundary state
-		prevId = getId();
-	}
+
 	lns.setBoundary(c.isBoundary); // update must be here
 
-	c.id = prevId;
 	if (isFork)
 		lns.setFork(sha);
 	if (isMerge)
@@ -779,7 +774,7 @@ void Git::updateLanes(Rev& c, Lanes& lns) {
 // 		tmp2.setNum(c.lanes[i]);
 // 		tmp.append(tmp2 + "-");
 // 	}
-// 	qDebug("%s  %s",tmp.latin1(), c.sha.latin1());
+// 	qDebug("%s %s",tmp.latin1(), c.sha.latin1());
 }
 
 void Git::on_procDataReady(const QString& fileChunk) {
@@ -900,6 +895,184 @@ void Git::updateRefs(Rev& c, bool addToHistory) {
 	}
 }
 
+void Git::updateDescMap(const Rev* r,uint idx, QMap<QPair<uint, uint>, bool>& dm,
+                        QMap<uint, QValueVector<int> >& dv) {
+
+	QValueVector<int> descVec(1, idx);
+
+	if (r->descRefsMaster != -1) {
+
+		const QValueVector<int>& nr = revLookup(revOrder[r->descRefsMaster])->descRefs;
+
+		for (uint i = 0; i < nr.count(); i++) {
+
+			if (!dv.contains(nr[i])) {
+				dbp("ASSERT descendant for %1 not found", r->sha());
+				return;
+			}
+			const QValueVector<int>& dvv = dv[nr[i]];
+			for (uint y = 0; y < dvv.count(); y++) {
+
+				QPair<uint, uint> key = qMakePair(idx, (uint)dvv[y]);
+				QPair<uint, uint> keyN = qMakePair((uint)dvv[y], idx);
+				dm.insert(key, true);
+				dm.insert(keyN, false);
+				descVec.append(dvv[y]);
+			}
+		}
+	}
+	dv.insert(idx, descVec);
+}
+
+void Git::mergeBranches(Rev* p, const Rev* r) {
+
+	int r_descBrnMaster = r->isBranch ? r->orderIdx : r->descBrnMaster;
+
+	if (p->descBrnMaster == r_descBrnMaster || r_descBrnMaster == -1)
+		return;
+
+	// we want all the descendant branches, so just avoid duplicates
+	const QValueVector<int>& src1 = revLookup(revOrder[p->descBrnMaster])->descBranches;
+	const QValueVector<int>& src2 = revLookup(revOrder[r_descBrnMaster])->descBranches;
+	QValueVector<int> dst(src1);
+	for (uint i = 0; i < src2.count(); i++)
+		if (qFind(src1.constBegin(), src1.constEnd(), src2[i]) == src1.constEnd())
+			dst.append(src2[i]);
+
+	p->descBranches = dst;
+	p->descBrnMaster = p->orderIdx;
+}
+
+void Git::mergeNearTags(bool down, Rev* p, const Rev* r, const QMap<QPair<uint, uint>, bool>& dm) {
+
+	int r_descRefsMaster = r->isTag ? r->orderIdx : r->descRefsMaster;
+	int r_ancRefsMaster = r->isTag ? r->orderIdx : r->ancRefsMaster;
+
+	if (down && (p->descRefsMaster == r_descRefsMaster || r_descRefsMaster == -1))
+		return;
+
+	if (!down && (p->ancRefsMaster == r_ancRefsMaster || r_ancRefsMaster == -1))
+		return;
+
+	if (dm.isEmpty()) {
+		dbs("ASSERT descendantMap is empty");
+		return;
+	}
+	// we want the nearest tag only, so remove any tag
+	// that is ancestor of any other tag in p U r
+	SCRef sha1 = down ? revOrder[p->descRefsMaster] : revOrder[p->ancRefsMaster];
+	SCRef sha2 = down ? revOrder[r_descRefsMaster] : revOrder[r_ancRefsMaster];
+	const QValueVector<int>& src1 = down ? revLookup(sha1)->descRefs : revLookup(sha1)->ancRefs;
+	const QValueVector<int>& src2 = down ? revLookup(sha2)->descRefs : revLookup(sha2)->ancRefs;
+	QValueVector<int> dst(src1);
+
+	for (uint s2 = 0; s2 < src2.count(); s2++) {
+
+		bool add = false;
+		for (uint s1 = 0; s1 < src1.count(); s1++) {
+
+			if (src2[s2] == src1[s1]) {
+				add = false;
+				break;
+			}
+			QPair<uint, uint> key = qMakePair((uint)src2[s2], (uint)src1[s1]);
+
+			if (!dm.contains(key)) {
+				add = true; // could be an indipendent path
+				continue;
+			}
+			add = (down && dm[key]) || (!down && !dm[key]);
+			if (add)
+				dst[s1] = -1; // mark for removing
+			else
+				break;
+		}
+		if (add)
+			dst.append(src2[s2]);
+	}
+	QValueVector<int>& nearRefs = (down) ? p->descRefs : p->ancRefs;
+	int& nearRefsMaster = (down) ? p->descRefsMaster : p->ancRefsMaster;
+
+	nearRefs.clear();
+	for (uint s2 = 0; s2 < dst.count(); s2++)
+		if (dst[s2] != -1)
+			nearRefs.append(dst[s2]);
+
+	nearRefsMaster = p->orderIdx;
+}
+
+void Git::indexTree() {
+
+	if (revOrder.count() == 0)
+		return;
+
+	// we keep the pairs(x, y). Value is true if x is
+	// ancestor of y or false if y is ancestor of x
+	QMap<QPair<uint, uint>, bool> descMap;
+	QMap<uint, QValueVector<int> > descVect;
+
+	// walk down the tree, compute chidlren and nearest descendants
+	for (uint i = 0; i < revOrder.count(); i++) {
+
+		const Rev* r = revLookup(revOrder[i]);
+
+		if (r->isBranch) {
+			Rev* rr = const_cast<Rev*>(r);
+			if (r->descBrnMaster != -1) {
+				SCRef sha = revOrder[r->descBrnMaster];
+				rr->descBranches = revLookup(sha)->descBranches;
+			}
+			rr->descBranches.append(i);
+		}
+		if (r->isTag) {
+			updateDescMap(r, i, descMap, descVect);
+			Rev* rr = const_cast<Rev*>(r);
+			rr->descRefs.clear();
+			rr->descRefs.append(i);
+		}
+		for (uint y = 0; y < r->parentsCount(); y++) {
+
+			Rev* p = const_cast<Rev*>(revLookup(r->parent(y)));
+			if (p) {
+				p->childs.append(i);
+
+				if (p->descBrnMaster == -1)
+					p->descBrnMaster = r->isBranch ? r->orderIdx :
+					                                 r->descBrnMaster;
+				else
+					mergeBranches(p, r);
+
+				if (p->descRefsMaster == -1)
+					p->descRefsMaster = r->isTag ? r->orderIdx :
+					                               r->descRefsMaster;
+				else
+					mergeNearTags(true, p, r, descMap);
+			}
+		}
+	}
+	// walk backward through the tree and compute nearest tagged ancestors
+	for (int i = revOrder.count() - 1; i >= 0 ; i--) {
+
+		const Rev* r = revLookup(revOrder[i]);
+
+		if (r->isTag) {
+			Rev* rr = const_cast<Rev*>(r);
+			rr->ancRefs.clear();
+			rr->ancRefs.append(i);
+		}
+		for (uint y = 0; y < r->childs.count(); y++) {
+
+			Rev* c = const_cast<Rev*>(revLookup(revOrder[r->childs[y]]));
+			if (c) {
+				if (c->ancRefsMaster == -1)
+					c->ancRefsMaster = r->isTag ? r->orderIdx:r->ancRefsMaster;
+				else
+					mergeNearTags(false, c, r, descMap);
+			}
+		}
+	}
+}
+
 // ********************************* Rev **************************
 
 const QString Rev::parent(int idx) const {
@@ -907,16 +1080,18 @@ const QString Rev::parent(int idx) const {
 	int start = 41 + 41 * idx;
 	if (isBoundary)
 		start++;
+
 	return d.mid(start, 40);
-};
+}
 
 const QStringList Rev::parents() const {
 
 	if (parentsCnt == 0)
 		return QStringList();
+
 	int start = (isBoundary) ? 42 : 41;
 	return QStringList::split(" ", d.mid(start, 41 * parentsCnt - 1));
-};
+}
 
 void Rev::indexData() const {
 
