@@ -23,7 +23,7 @@
 using namespace QGit;
 
 ListView::ListView(Domain* dm, Git* g, QListView* l, bool h) :
-		QObject(dm), d(dm), git(g), lv(l), isHistory(h) {
+                   QObject(dm), d(dm), git(g), lv(l), isHistory(h) {
 
 	st = &(d->st);
 	processedRevs = 0;
@@ -41,17 +41,17 @@ ListView::ListView(Domain* dm, Git* g, QListView* l, bool h) :
 		lv->setColumnText(TIME_COL, "Author Date");
 
 	connect(lv, SIGNAL(currentChanged(QListViewItem*)),
-		this, SLOT(on_currentChanged(QListViewItem*)));
+	        this, SLOT(on_currentChanged(QListViewItem*)));
 	connect(lv, SIGNAL(contentsMoving(int,int) ),
-		this, SLOT(on_contentsMoving(int,int)));
+	        this, SLOT(on_contentsMoving(int,int)));
 	connect(lv, SIGNAL(mouseButtonPressed(int,QListViewItem*,const QPoint&,int)),
-		this, SLOT(on_mouseButtonPressed(int,QListViewItem*,const QPoint&,int)));
+	        this, SLOT(on_mouseButtonPressed(int,QListViewItem*,const QPoint&,int)));
 	connect(lv, SIGNAL(clicked(QListViewItem*)),
-		this, SLOT(on_clicked(QListViewItem*)));
+	        this, SLOT(on_clicked(QListViewItem*)));
 	connect(lv, SIGNAL(onItem(QListViewItem*)),
-		this, SLOT(on_onItem(QListViewItem*)));
+	        this, SLOT(on_onItem(QListViewItem*)));
 	connect(lv, SIGNAL(contextMenuRequested(QListViewItem*, const QPoint&,int)),
-		this, SLOT(on_contextMenuRequested(QListViewItem*)));
+	        this, SLOT(on_contextMenuRequested(QListViewItem*)));
 }
 
 ListView::~ListView() {
@@ -210,7 +210,7 @@ void ListView::on_newRevsAdded(const QValueVector<QString>& shaVec) {
 
 	for (uint i = processedRevs; i < shaVec.count(); ++i) {
 		lastItem = new ListViewItem(lv, lastItem, git, shaVec[i], d->m()->pixmaps,
-							   evenLine, d->m()->secs, isHistory);
+		                            evenLine, d->m()->secs, isHistory);
 		evenLine = !evenLine;
 	}
 	processedRevs = shaVec.count();
@@ -245,17 +245,17 @@ void ListView::on_onItem(QListViewItem*) {
 	if (d->isReadyToDrag()) {
 		if (!d->setDragging(true))
 			return;
+
 		QStringList dragRevs;
 		getSelectedItems(dragRevs);
 		if (!dragRevs.contains(ZERO_SHA)) {
+
 			const QString host(QString::fromLatin1("@") + d->m()->curDir);
 			for (uint i = 0; i < dragRevs.count(); i++)
 				dragRevs[i].append(host);
 
-			lv->viewport()->setAcceptDrops(false);
 			QDragObject* drObj = new QTextDrag(dragRevs.join("\n"), lv);
 			drObj->dragCopy(); // do NOT delete drObj. Blocking until drop event
-			lv->viewport()->setAcceptDrops(!isHistory);
 		}
 		d->setDragging(false);
 	}
@@ -335,10 +335,13 @@ bool ListView::filterRightButtonPressed(QMouseEvent* e) {
 
 		SCRef diffToSha(item->text(COMMIT_COL));
 
-		if (diffToSha != ZERO_SHA && st->sha() != ZERO_SHA &&
-			diffToSha != st->diffToSha()) {
+		if (diffToSha != ZERO_SHA && st->sha() != ZERO_SHA) {
+
+			if (diffToSha != st->diffToSha())
+				st->setDiffToSha(diffToSha);
+			else
+				st->setDiffToSha(""); // restore std view
 
-			st->setDiffToSha(diffToSha);
 			UPDATE_DOMAIN(d);
 			filterNextContextMenuRequest = true;
 			return true; // filter event out
@@ -407,7 +410,7 @@ ListViewItem::ListViewItem(QListView* p, ListViewItem* a, Git* g, SCRef sha,
 	isEvenLine = e;
 	secs = t;
 	isHistory = h;
-	populated = isDiffTarget = false;
+	populated = isDiffTarget = isHighlighted = false;
 	setText(COMMIT_COL, sha); // item key, must be set from beginning
 }
 
@@ -424,10 +427,8 @@ void ListViewItem::setDiffTarget(bool b) {
 }
 
 void ListViewItem::paintCell(QPainter* p, const QColorGroup& cg,
-						  int column, int width, int alignment) {
+                             int column, int width, int alignment) {
 	QColorGroup _cg(cg);
-	QColor c = _cg.base();
-
 	const Rev& r = *git->revLookup(text(COMMIT_COL), isHistory);
 
 	// lazy setup, only when visible
@@ -435,7 +436,7 @@ void ListViewItem::paintCell(QPainter* p, const QColorGroup& cg,
 		setupData(r);
 
 	// pixmap graph, separated from setupData to allow deleting
-	if (pixmap(GRAPH_COL) == NULL) {
+	if (!pixmap(GRAPH_COL)) {
 		QPixmap* pm = getGraph(r);
 		setPixmap(GRAPH_COL, *pm);
 		delete pm;
@@ -453,18 +454,13 @@ void ListViewItem::paintCell(QPainter* p, const QColorGroup& cg,
 
 	// tags, heads, refs and working dir colouring
 	if (mycolumn == LOG_COL) {
-		if (r.isTag)
-			_cg.setColor(QColorGroup::Base, Qt::yellow);
-
-		if (r.isBranch)
-			_cg.setColor(QColorGroup::Base, DARK_GREEN);
-
-		if (r.isCurrentBranch)
-			_cg.setColor(QColorGroup::Base, Qt::green);
-
-		if (r.isRef)
-			_cg.setColor(QColorGroup::Base, PURPLE);
 
+		if (isHighlighted) {
+			QFont f(p->font());
+			f.setBold(true);
+			p->save();
+			p->setFont(f);
+		}
 		if (r.isDiffCache) {
 			if (changedFiles(ZERO_SHA))
 				_cg.setColor(QColorGroup::Base, ORANGE);
@@ -477,7 +473,47 @@ void ListViewItem::paintCell(QPainter* p, const QColorGroup& cg,
 		_cg.setColor(QColorGroup::Base, LIGHT_BLUE);
 
 	QListViewItem::paintCell(p, _cg, column, width, alignment);
-	_cg.setColor(QColorGroup::Base, c);
+
+	if (isHighlighted && mycolumn == LOG_COL)
+		p->restore();
+}
+
+void ListViewItem::addTextPixmap(SCRef text, const QColor& color) {
+
+	int col = LOG_COL + ((isHistory) ? 0 : -1);
+	QStringList sl = QStringList::split('\n', text);
+	loopList(it, sl) {
+		QPixmap* pm = doAddTextPixmap(*it, color, col);
+		setPixmap(col, *pm);
+		delete pm;
+	}
+}
+
+QPixmap* ListViewItem::doAddTextPixmap(SCRef text, const QColor& color, int col) {
+
+	const QPixmap* oldPm = pixmap(col);
+	QFontMetrics fm(listView()->font());
+	int spacing = 2;
+	int pw = fm.boundingRect(text).width() + 2 * spacing;
+	int ph = fm.height() + 1;
+	int ofs = oldPm ? oldPm->width() + 2 : 0;
+
+	QPixmap* pm = new QPixmap(ofs + pw, ph);
+
+	QPainter p;
+	p.begin(pm);
+
+	if (oldPm) {
+		pm->fill(isEvenLine ? EVEN_LINE_COL : ODD_LINE_COL);
+		p.drawPixmap(0, 0, *oldPm);
+	}
+	p.setPen(Qt::black);
+	p.setBrush(color);
+	p.setFont(listView()->font());
+	p.drawRect(ofs, 0, pw, ph);
+	p.drawText(ofs + spacing, fm.ascent(), text);
+	p.end();
+	return pm;
 }
 
 bool ListViewItem::changedFiles(SCRef c) {
@@ -511,6 +547,15 @@ void ListViewItem::setupData(const Rev& c) {
 	}
 	setText(LOG_COL + adj, c.shortLog());
 	setText(AUTH_COL + adj, c.author());
+
+	if (c.isBranch || c.isCurrentBranch)
+		addTextPixmap(c.branch, c.isCurrentBranch ? Qt::green : DARK_GREEN);
+
+	if (c.isTag)
+		addTextPixmap(c.tag, Qt::yellow);
+
+	if (c.isRef)
+		addTextPixmap(c.ref, PURPLE);
 }
 
 const QString ListViewItem::timeDiff(unsigned long secs) const {
@@ -601,7 +646,7 @@ void MainImpl::setupPixmaps(int h) {
 
 	QBrush myWhiteBrush(ODD_LINE_COL, Qt::SolidPattern);
 	QColor colors[COLORS_NUM] = {Qt::black, Qt::red, DARK_GREEN, Qt::blue,
-					Qt::darkGray, BROWN, Qt::magenta, ORANGE};
+	                             Qt::darkGray, BROWN, Qt::magenta, ORANGE};
 
 	pixmaps.clear();
 	pixmaps.resize(LANE_TYPES_NUM * COLORS_NUM);
