@@ -31,10 +31,15 @@
 #include <qaccel.h>
 #include <qsplitter.h>
 #include <qtabwidget.h>
+#include <qobjectlist.h>
+#include <qlayout.h>
 #include "config.h" // defines PACKAGE_VERSION
 #include "help.h"
 #include "helpbase.h"
+#include "consoleimpl.h"
+#include "macroimpl.h"
 #include "settingsimpl.h"
+#include "filebase.h"
 #include "common.h"
 #include "git.h"
 #include "listview.h"
@@ -43,6 +48,7 @@
 #include "fileview.h"
 #include "commitimpl.h"
 #include "revsview.h"
+#include "revdesc.h"
 #include "mainimpl.h"
 
 using namespace QGit;
@@ -52,26 +58,32 @@ using namespace QGit;
 QColor QGit::ODD_LINE_COL; // declared as extern in QGit
 QColor QGit::EVEN_LINE_COL;
 
-uint QGit::flags() {
+uint QGit::flags(SCRef group) {
 
 	QSettings settings;
-	return settings.readNumEntry(APP_KEY + FLAGS_KEY, FLAGS_DEF);
+	return settings.readNumEntry(APP_KEY + group + FLAGS_KEY, FLAGS_DEF);
 }
 
-bool QGit::testFlag(uint f) { return (flags() & f); }
+bool QGit::testFlag(uint f, SCRef group) { return (flags(group) & f); }
 
-void QGit::setFlag(uint f, bool b) {
+void QGit::setFlag(uint f, bool b, SCRef group) {
 
 	QSettings settings;
-	int flags = settings.readNumEntry(APP_KEY + FLAGS_KEY, FLAGS_DEF);
+	int flags = settings.readNumEntry(APP_KEY + group + FLAGS_KEY, FLAGS_DEF);
 	flags = (b) ? flags | f : flags & ~f;
-	settings.writeEntry(APP_KEY + FLAGS_KEY, flags);
+	settings.writeEntry(APP_KEY + group + FLAGS_KEY, flags);
+}
+
+void QGit::writeSetting(SCRef key, SCRef value, SCRef group) {
+
+	QSettings settings;
+	settings.writeEntry(APP_KEY + group + key, value);
 }
 
 // ******************************* MainImpl ******************************
 
-MainImpl::MainImpl(const QString& cd, QWidget* p, const char *name)
-		: MainBase(p, name, Qt::WDestructiveClose) {
+MainImpl::MainImpl(const QString& cd, QWidget* p, const char *name) :
+          MainBase(p, name, Qt::WDestructiveClose) {
 
 	EM_INIT(exSetRepositoryCalled, "Reloading data");
 	EM_INIT(exExiting, "Exiting");
@@ -91,11 +103,17 @@ MainImpl::MainImpl(const QString& cd, QWidget* p, const char *name)
 	longLogRE.setCaseSensitive(false);
 
 	// set-up menu for recent visited repositories
-	connect(File, SIGNAL(activated(int)), this, SLOT(openRecent_Activated(int)));
+	connect(File, SIGNAL(activated(int)), this, SLOT(on_openRecent_activated(int)));
 	recentRepoMenuPos = 0;
 	while (File->idAt(recentRepoMenuPos) != -1)
 		recentRepoMenuPos++;
-	updateRecentRepoMenu();
+	doUpdateRecentRepoMenu("");
+
+	// set-up menu for macros
+	connect(Actions, SIGNAL(activated(int)), this, SLOT(on_macro_activated(int)));
+	QSettings set;
+	QStringList sl = QStringList::split(",", set.readEntry(APP_KEY + MCR_LIST_KEY, ""));
+	doUpdateMacroMenu(sl);
 
 	// create light and dark colors for alternate background
 	QColor l(listViewLog->paletteBackgroundColor());
@@ -115,7 +133,9 @@ MainImpl::MainImpl(const QString& cd, QWidget* p, const char *name)
 	// set-up tab view
 	revTab = tabWdg->page(0);
 	patchTab = tabWdg->page(1);
-	fileTab = tabWdg->page(2);
+	fileTab = new TabFile(this);
+	tabWdg->addTab(fileTab, "File");
+
 	pushButtonCloseTab->reparent(tabWdg, QPoint());
 	tabWdg->setCornerWidget(pushButtonCloseTab);
 	tabWdg->removePage(patchTab);
@@ -127,23 +147,28 @@ MainImpl::MainImpl(const QString& cd, QWidget* p, const char *name)
 	adjustFontSize(0); // default is system-wide setting
 
 	connect(git, SIGNAL(newRevsAdded(const QValueVector<QString>&)),
-		this, SLOT(on_newRevsAdded()));
+	        this, SLOT(on_newRevsAdded()));
 
 	// connect cross-domain update signals
 	connect(listViewLog, SIGNAL(doubleClicked(QListViewItem*)),
-		   this, SLOT(on_listViewLog_doubleClicked(QListViewItem*)));
+             this, SLOT(on_listViewLog_doubleClicked(QListViewItem*)));
 	connect(listBoxFiles, SIGNAL(doubleClicked(QListBoxItem*)),
-		   this, SLOT(on_fileList_doubleClicked(QListBoxItem*)));
+             this, SLOT(on_fileList_doubleClicked(QListBoxItem*)));
 	connect(listBoxFiles_2, SIGNAL(doubleClicked(QListBoxItem*)),
-		   this, SLOT(on_fileList_doubleClicked(QListBoxItem*)));
+	        this, SLOT(on_fileList_doubleClicked(QListBoxItem*)));
 	connect(treeView, SIGNAL(doubleClicked(QListViewItem*)),
-		   this, SLOT(on_treeView_doubleClicked(QListViewItem*)));
-	connect(histListView, SIGNAL(doubleClicked(QListViewItem*)),
-		   this, SLOT(on_histListView_doubleClicked(QListViewItem*)));
+	        this, SLOT(on_treeView_doubleClicked(QListViewItem*)));
+	connect(fileTab->histListView, SIGNAL(doubleClicked(QListViewItem*)),
+	        this, SLOT(on_histListView_doubleClicked(QListViewItem*)));
 
-	// open repository outside c'tor
+	// some stuff must be done outside c'tor
 	startUpDir = (cd.isEmpty()) ? QDir::current().absPath() : cd;
-	QTimer::singleShot(10, this, SLOT(setRepositoryStartup()));
+	QTimer::singleShot(10, this, SLOT(afterCtorInit()));
+}
+
+void MainImpl::afterCtorInit() {
+
+	setRepository(startUpDir, false, false);
 }
 
 void MainImpl::buttonGroupDiff_clicked(int diffType) {
@@ -163,8 +188,8 @@ void MainImpl::ActBack_activated() {
 
 	lineEditSHA->undo(); // first for insert(text)
 	if (lineEditSHA->text().isEmpty())
-		// second for clear() if text has been modified by updateLineEditSHA()
-		lineEditSHA->undo();
+		lineEditSHA->undo(); // double undo, see RevsView::updateLineEditSHA()
+
 	lineEditSHA_returnPressed();
 }
 
@@ -228,10 +253,8 @@ void MainImpl::getExternalDiffArgs(QStringList* args) {
 
 // *************************** Repo open or changed *************************
 
-void MainImpl::setRepositoryStartup() { setRepository(startUpDir, false, false); }
-
 void MainImpl::setRepository(SCRef newDir, bool refresh, bool keepSelection,
-					    QStringList* filterList) {
+                             QStringList* filterList) {
 
 	if (refresh && ActFilterTree->isOn() && filterList == NULL) {
 		// refresh called while in filtered view
@@ -281,7 +304,7 @@ void MainImpl::setRepository(SCRef newDir, bool refresh, bool keepSelection,
 		}
 		if (ActFilterTree->isOn())
 			setCaption(caption() + " - FILTER ON < " +
-					filterList->join(" ") + " >");
+			           filterList->join(" ") + " >");
 
 		// tree name should be set before init because in case of
 		// StGIT archives the first revs are sent before init returns
@@ -319,11 +342,12 @@ void MainImpl::setRepository(SCRef newDir, bool refresh, bool keepSelection,
 			return;
 		}
 		const QString info("Exception \'" + EM_DESC(i) + "\' "
-				"not handled in setRepository...re-throw");
+		                   "not handled in setRepository...re-throw");
 		dbp("%1", info.latin1());
 		throw i;
 	}
 }
+
 void MainImpl::updateGlobalActions(bool b) {
 
 	ActRefresh->setEnabled(b);
@@ -339,7 +363,7 @@ void MainImpl::updateGlobalActions(bool b) {
 }
 
 void MainImpl::updateContextActions(SCRef newRevSha, SCRef newFileName,
-							bool isDir, bool found) {
+                                    bool isDir, bool found) {
 
 	bool pathActionsEnabled = !newFileName.isEmpty();
 	bool fileActionsEnabled = pathActionsEnabled && !isDir;
@@ -380,12 +404,15 @@ void MainImpl::on_histListView_doubleClicked(QListViewItem* item) {
 
 void MainImpl::on_fileList_doubleClicked(QListBoxItem* item) {
 
-	if (rv->st.isMerge() &&
-		(item == listBoxFiles->firstItem() ||item == listBoxFiles_2->firstItem()))
+	if (rv->st.isMerge() && (item == listBoxFiles->firstItem() ||
+	    item == listBoxFiles_2->firstItem()))
 		return;
 
-	if (item && ActViewDiff->isEnabled())
+	if (item && item->listBox() == listBoxFiles && ActViewDiff->isEnabled())
 		ActViewDiff->activate();
+
+	if (item && item->listBox() == listBoxFiles_2 && ActViewFile->isEnabled())
+		ActViewFile->activate();
 }
 
 void MainImpl::on_treeView_doubleClicked(QListViewItem* item) {
@@ -473,9 +500,12 @@ void MainImpl::on_newRevsAdded() {
 	if (toolButtonFilter->isOn())
 		toolButtonFilter_toggled(true); // filter again on new arrived data
 
+	if (toolButtonBold->isOn())
+		toolButtonBold_toggled(true); // filter again on new arrived data
+
 	// first rev could be a StGIT unapplied patch so check more then once
 	if (!ActCommit->isEnabled() && !git->isCommittingMerge() &&
-		(!git->isNothingToCommit() || git->isUnknownFiles()))
+	   (!git->isNothingToCommit() || git->isUnknownFiles()))
 		ActCommit_setEnabled(true);
 }
 
@@ -486,6 +516,18 @@ void MainImpl::lineEditFilter_returnPressed() {
 
 void MainImpl::toolButtonFilter_toggled(bool isOn) {
 
+	toolButtonBold->setEnabled(!isOn);
+	filterList(isOn, false);
+}
+
+void MainImpl::toolButtonBold_toggled(bool isOn) {
+
+	toolButtonFilter->setEnabled(!isOn);
+	filterList(isOn, true);
+}
+
+void MainImpl::filterList(bool isOn, bool onlyHighlight) {
+
 	lineEditFilter->setEnabled(!isOn);
 	cmbSearch->setEnabled(!isOn);
 
@@ -542,12 +584,17 @@ void MainImpl::toolButtonFilter_toggled(bool isOn) {
 		ListViewItem* item = static_cast<ListViewItem*>(it.current());
 		if (isOn) {
 			if (passFilter(item, filter, colNum, shaMap)) {
-				item->setEven(evenLine);
-				evenLine = !evenLine;
+				if (onlyHighlight)
+					item->setHighlighted(true);
+				else {
+					item->setEven(evenLine);
+					evenLine = !evenLine;
+				}
 				visibleCnt++;
-			} else
+			} else if (!onlyHighlight)
 				item->setVisible(false);
 		} else {
+			item->setHighlighted(false);
 			item->setEven(evenLine);
 			evenLine = !evenLine;
 			if (!item->isVisible())
@@ -562,17 +609,19 @@ void MainImpl::toolButtonFilter_toggled(bool isOn) {
 		emit highlightPatch(isOn ? filter : "", isRegExp);
 
 	if (isOn) {
-		const QString tmp(QString("Found %1 matches. Toggle filter "
-				"button to remove the filter").arg(visibleCnt));
+		const QString tmp(QString("Found %1 matches. Toggle filter/highlight "
+		                          "button to remove the filter").arg(visibleCnt));
 		statusBar()->message(tmp);
 	} else {
 		listViewLog->ensureItemVisible(listViewLog->currentItem());
 		statusBar()->clear();
 	}
+	scrollListView(1); // hack to force a repaint
+	scrollListView(-1);
 }
 
-bool MainImpl::passFilter(ListViewItem* item, SCRef filter,
-					 int colNum, const QMap<QString, bool>& shaMap) {
+bool MainImpl::passFilter(ListViewItem* item, SCRef filter, int colNum,
+                          const QMap<QString, bool>& shaMap) {
 	QString field;
 	if (colNum == SHA_MAP_COL) {
 		// in this case shaMap contains all good sha to search for
@@ -653,7 +702,7 @@ void MainImpl::tabWdg_currentChanged(QWidget*) {
 		textEditDiff->setFocus();
 		break;
 	case TAB_FILE:
-		histListView->setFocus();
+		fileTab->histListView->setFocus();
 		break;
 	default:
 		dbs("ASSERT in tabWdg_currentChanged: unknown current page");
@@ -670,6 +719,12 @@ void MainImpl::accelActivated(int id) {
 	case KEY_DOWN:
 		scrollListView(1);
 		break;
+	case SHIFT_KEY_UP:
+		goMatch(-1);
+		break;
+	case SHIFT_KEY_DOWN:
+		goMatch(1);
+		break;
 	case KEY_LEFT:
 		ActBack_activated();
 		break;
@@ -713,6 +768,8 @@ void MainImpl::setupAccelerator(QAccel* accel) {
 	accel->insertItem(Key_Up, KEY_UP);
 	accel->insertItem(Key_I, KEY_UP);
 	accel->insertItem(Key_Down, KEY_DOWN);
+	accel->insertItem(SHIFT+Key_Up, SHIFT_KEY_UP);
+	accel->insertItem(SHIFT+Key_Down, SHIFT_KEY_DOWN);
 	accel->insertItem(Key_N, KEY_DOWN);
 	accel->insertItem(Key_K, KEY_DOWN);
 	accel->insertItem(Key_Left, KEY_LEFT);
@@ -732,6 +789,31 @@ void MainImpl::setupAccelerator(QAccel* accel) {
 	connect(accel, SIGNAL(activated(int)), this, SLOT(accelActivated(int)));
 }
 
+void MainImpl::goMatch(int delta) {
+
+	if (!toolButtonBold->isOn())
+		return;
+
+	QListViewItemIterator it(listViewLog->currentItem());
+	if (delta > 0)
+		++it;
+	else
+		--it;
+	while (it.current()) {
+		ListViewItem* item = static_cast<ListViewItem*>(it.current());
+		if (item->highlighted()) {
+			listViewLog->clearSelection();
+			listViewLog->setCurrentItem(item);
+			listViewLog->ensureItemVisible(listViewLog->currentItem());
+			return;
+		}
+		if (delta > 0)
+			++it;
+		else
+			--it;
+	}
+}
+
 QTextEdit* MainImpl::getCurrentTextEdit() {
 
 	QTextEdit* te = NULL;
@@ -743,7 +825,7 @@ QTextEdit* MainImpl::getCurrentTextEdit() {
 		te = textEditDiff;
 		break;
 	case TAB_FILE:
-		te = textEditFile;
+		te = fileTab->textEditFile;
 		break;
 	default:
 		break;
@@ -775,7 +857,7 @@ void MainImpl::scrollListView(int delta) {
 		lv = listViewLog;
 		break;
 	case TAB_FILE:
-		lv = histListView;
+		lv = fileTab->histListView;
 		break;
 	default:
 		lv = qApp->focusWidget();
@@ -867,7 +949,7 @@ void MainImpl::doUpdateRecentRepoMenu(SCRef newEntry) {
 		File->insertItem(QString::number(i) + " " + *it2);
 
 	for (int i = recents.count() - MAX_RECENT_REPOS; i > 0; i--)
-			recents.pop_back();
+		recents.pop_back();
 
 	settings.writeEntry(APP_KEY + REC_REP_KEY, recents.join(","));
 }
@@ -883,8 +965,8 @@ void MainImpl::doContexPopup(SCRef sha) {
 
 	contextMenu = new QPopupMenu();
 	contextSubMenu = new QPopupMenu();
-	connect(contextMenu, SIGNAL(activated(int)), this, SLOT(goRef_Activated(int)));
-	connect(contextSubMenu, SIGNAL(activated(int)), this, SLOT(goRef_Activated(int)));
+	connect(contextMenu, SIGNAL(activated(int)), this, SLOT(on_goRef_activated(int)));
+	connect(contextSubMenu, SIGNAL(activated(int)), this, SLOT(on_goRef_activated(int)));
 
 	bool isRevPage = (tabWdg->currentPageIndex() == tabWdg->indexOf(revTab));
 	bool isPatchPage = (tabWdg->currentPageIndex() == tabWdg->indexOf(patchTab));
@@ -985,7 +1067,7 @@ void MainImpl::doFileContexPopup(SCRef fileName, int type) {
 	contextMenu.exec(QCursor::pos());
 }
 
-void MainImpl::goRef_Activated(int id) {
+void MainImpl::on_goRef_activated(int id) {
 
 	if (id <= 0) // not a tag name entry
 		return;
@@ -1009,8 +1091,8 @@ void MainImpl::ActSplitView_activated() {
 		textBrowserDesc_2->setHidden(hide);
 		break;
 	case TAB_FILE:
-		hide = histListView->isVisible();
-		histListView->setHidden(hide);
+		hide = fileTab->histListView->isVisible();
+		fileTab->histListView->setHidden(hide);
 		break;
 	default:
 		dbs("ASSERT in ActSplitView_activated: unknown current page");
@@ -1030,9 +1112,8 @@ void MainImpl::ActShowTree_toggled(bool b) {
 void MainImpl::ActSaveFile_activated() {
 
 	QFileInfo f(rv->st.fileName());
-	const QString fileName(QFileDialog::getSaveFileName(
-			f.fileName(), QString::null, this,
-			"save file dialog", "Save file as"));
+	const QString fileName(QFileDialog::getSaveFileName(f.fileName(), "",
+	                       this, "save file dialog", "Save file as"));
 
 	if (fileName.isEmpty())
 		return;
@@ -1048,7 +1129,7 @@ void MainImpl::ActSaveFile_activated() {
 		statusBar()->message("Unable to save " + fileName);
 }
 
-void MainImpl::openRecent_Activated(int id) {
+void MainImpl::on_openRecent_activated(int id) {
 
 	bool ok;
 	File->text(id).left(1).toInt(&ok);
@@ -1062,8 +1143,8 @@ void MainImpl::openRecent_Activated(int id) {
 
 void MainImpl::ActOpenRepo_activated() {
 
-	const QString dirName(QFileDialog::getExistingDirectory(
-			curDir, this, "", "Choose a directory"));
+	const QString dirName(QFileDialog::getExistingDirectory(curDir,
+	                      this, "", "Choose a directory"));
 
 	if (!dirName.isEmpty()) {
 		QDir d(dirName);
@@ -1073,8 +1154,8 @@ void MainImpl::ActOpenRepo_activated() {
 
 void MainImpl::ActOpenRepoNewWindow_activated() {
 
-	const QString dirName(QFileDialog::getExistingDirectory(
-			curDir, this, "", "Choose a directory"));
+	const QString dirName(QFileDialog::getExistingDirectory(curDir,
+	                      this, "", "Choose a directory"));
 
 	if (!dirName.isEmpty()) {
 		QDir d(dirName);
@@ -1111,7 +1192,7 @@ void MainImpl::ActMailFormatPatch_activated() {
 	QSettings settings;
 	QString outDir(settings.readEntry(APP_KEY + FP_DIR_KEY, curDir));
 	QString dirPath(QFileDialog::getExistingDirectory(outDir, this, "",
-		"Choose destination directory - Format Patch"));
+	                "Choose destination directory - Format Patch"));
 	if (dirPath.isEmpty())
 		return;
 
@@ -1122,30 +1203,50 @@ void MainImpl::ActMailFormatPatch_activated() {
 	QApplication::restoreOverrideCursor();
 }
 
+bool MainImpl::askApplyPatchParameters(bool* commit, bool* fold) {
+
+	int ret = QMessageBox::question(this, "Apply Patch",
+	          "Do you want to commit or just to apply changes to "
+	          "working directory?", "&Cancel", "&Working dir", "&Commit", 0, 0);
+	if (ret == 0)
+		return false;
+
+	*commit = (ret == 2);
+	*fold = false;
+	if (*commit && git->isStGITStack()) {
+		ret = QMessageBox::question(this, "Apply Patch", "Do you want to "
+		      "import or fold the patch?", "&Cancel", "&Fold", "&Import", 0, 0);
+		if (ret == 0)
+			return false;
+
+		*fold = (ret == 1);
+	}
+	return true;
+}
+
 void MainImpl::ActMailApplyPatch_activated() {
 
-	bool fold = false;
 	QSettings settings;
 	QString outDir(settings.readEntry(APP_KEY + FP_DIR_KEY, curDir));
-	QString patchName(QFileDialog::getOpenFileName(outDir, NULL,
-			this, "", "Choose the patch file - Apply Patch"));
+	QString patchName(QFileDialog::getOpenFileName(outDir, NULL, this,
+	                  "", "Choose the patch file - Apply Patch"));
 	if (patchName.isEmpty())
 		return;
 
 	QFileInfo f(patchName);
 	settings.writeEntry(APP_KEY + FP_DIR_KEY, f.dirPath(true));
-	if (git->isStGITStack()) {
-		int ret = QMessageBox::question(this, "Apply Patch",
-			"Do you want to import or fold the patch?",
-			"&Cancel", "&Fold", "&Import", 0, 0);
-		if (ret == 0)
-			return;
-		fold = (ret == 1);
-	}
+
+	bool commit, fold;
+	if (!askApplyPatchParameters(&commit, &fold))
+		return;
+
 	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
-	git->applyPatchFile(f.absFilePath(), false, fold);
+
+	if (git->applyPatchFile(f.absFilePath(), commit, fold, false) && !commit)
+		git->resetCommits(1);
+
 	QApplication::restoreOverrideCursor();
-	refreshRepo();
+	refreshRepo(false);
 }
 
 void MainImpl::ActCheckWorkDir_toggled(bool b) {
@@ -1161,7 +1262,7 @@ void MainImpl::ActCheckWorkDir_toggled(bool b) {
 void MainImpl::ActSettings_activated() {
 
 	SettingsImpl* setView = new SettingsImpl(this, git);
-	setView->exec();
+	setView->exec(); // modal exec
 
 	// update ActCheckWorkDir if necessary
 	bool diffIndex = (testFlag(DIFF_INDEX_F) != 0);
@@ -1169,10 +1270,92 @@ void MainImpl::ActSettings_activated() {
 		ActCheckWorkDir->toggle();
 }
 
+void MainImpl::ActMacroSetup_activated() {
+
+	MacroImpl* macroView = new MacroImpl(); // has WDestructiveClose
+	connect(this, SIGNAL(closeAllWindows()), macroView, SLOT(close()));
+	connect(macroView, SIGNAL(macroListChanged(const QStringList&)),
+	        this, SLOT(on_macroListChanged(const QStringList&)));
+
+	macroView->show();
+}
+
+void MainImpl::on_macroListChanged(const QStringList& macroList) {
+
+	// update menu of all windows
+	QWidgetList* list = QApplication::topLevelWidgets();
+	QWidgetListIt it(*list);
+	while (it.current() != 0) {
+		MainImpl* w = dynamic_cast<MainImpl*>(it.current());
+		if (w)
+			w->doUpdateMacroMenu(macroList);
+		++it;
+	}
+	delete list; // list is created using new by topLevelWidgets()
+}
+
+void MainImpl::doUpdateMacroMenu(const QStringList& macroList) {
+
+	while (Actions->idAt(1) != -1) // clears menu
+		Actions->removeItemAt(1);
+
+	if (macroList.isEmpty())
+		return;
+
+	Actions->insertSeparator();
+	loopList(it, macroList)
+		Actions->insertItem(*it);
+}
+
+void MainImpl::on_macro_activated(int id) {
+
+	const QString name = Actions->text(id);
+	if (name == "Setup actions...")
+		return;
+
+	QSettings set;
+	QStringList sl = QStringList::split(",", set.readEntry(APP_KEY + MCR_LIST_KEY, ""));
+
+	if (sl.findIndex(name) == -1) {
+		dbp("ASSERT in on_macro_Activated, action %1 not found", name);
+		return;
+	}
+	const QString header = "Macro " + name + "/";
+	QString cmdArgs;
+
+	if (testFlag(MCR_CMD_LINE_F, header)) {
+		bool ok;
+		cmdArgs = QInputDialog::getText("Run action - QGit", "Enter command line "
+		          "arguments for '" + name + "'", QLineEdit::Normal, "", &ok, this);
+		cmdArgs.prepend(' ');
+		if (!ok)
+			return;
+	}
+	QString cmd = set.readEntry(APP_KEY + header + MCR_TEXT_KEY, "");
+	if (cmd.isEmpty())
+		return;
+
+	ConsoleImpl* c = new ConsoleImpl(name, git); // has WDestructiveClose
+	connect(this, SIGNAL(closeAllWindows()), c, SLOT(close()));
+	connect(c, SIGNAL(macro_exited(const QString&)),
+	        this, SLOT(on_macro_exited(const QString&)));
+
+	if (c->start(cmd, cmdArgs))
+		c->show();
+}
+
+void MainImpl::on_macro_exited(const QString& name) {
+
+	const QString header = "Macro " + name + "/";
+	if (testFlag(MCR_REFRESH_F, header))
+		QTimer::singleShot(10, this, SLOT(refreshRepo())); // must be out of handler
+}
+
 void MainImpl::ActCommit_activated() {
 
 	ActCommit_setEnabled(false);
-	commitViewer = new CommitImpl(git, &changesCommitted, this);
+	CommitImpl* commitViewer = new CommitImpl(git, &changesCommitted); // has WDestructiveClose
+	connect(this, SIGNAL(closeAllWindows()), commitViewer, SLOT(close()));
 	connect(commitViewer, SIGNAL(destroyed()), this, SLOT(slotCommitViewerClosed()));
 	commitViewer->show();
 }
@@ -1192,8 +1375,8 @@ void MainImpl::slotCommitViewerClosed() {
 
 	if (changesCommitted) {
 		changesCommitted = false;
-		refreshRepo();
-	} else  {
+		refreshRepo(false);
+	} else {
 		QListViewItem* item = listViewLog->findItem(ZERO_SHA, COMMIT_COL);
 		ActCommit_setEnabled(item != NULL && !git->isCommittingMerge());
 		statusBar()->message("Failed to commit changes");
@@ -1205,49 +1388,49 @@ void MainImpl::ActTag_activated() {
 	bool ok;
 	int adj = -1; // hack to correctly map col numbers in main view
 	QString tag(listViewLog->currentItem()->text(LOG_COL + adj));
-	tag = QInputDialog::getText("Make tag - QGit",
-		"Enter tag name:", QLineEdit::Normal, tag, &ok, this);
+	tag = QInputDialog::getText("Make tag - QGit", "Enter tag name:",
+	                            QLineEdit::Normal, tag, &ok, this);
 	if (!ok || tag.isEmpty())
 		return;
 	QString tmp(tag.simplifyWhiteSpace());
 	if (tag != tmp.remove(' ')) {
 		QMessageBox::warning(this, "Make tag - QGit",
-			"Sorry, control characters or spaces\n"
-			"are not allowed in tag name.");
+		             "Sorry, control characters or spaces\n"
+		             "are not allowed in tag name.");
 		return;
 	}
 	if (git->isTagName(tag)) {
 		QMessageBox::warning(this, "Make tag - QGit",
-			"Sorry, tag name already exists.\n"
-			"Please choose a different name.");
+		             "Sorry, tag name already exists.\n"
+		             "Please choose a different name.");
 		return;
 	}
 	QString msg(QInputDialog::getText("Make tag - QGit",
-		"Enter tag message, if any:", QLineEdit::Normal, "", &ok, this));
+	        "Enter tag message, if any:", QLineEdit::Normal, "", &ok, this));
 
 	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
 	ok = git->makeTag(lineEditSHA->text(), tag, msg);
 	QApplication::restoreOverrideCursor();
-	if (!ok)
-		statusBar()->message("Sorry, unable to tag the revision");
-	else
+	if (ok)
 		refreshRepo(true);
+	else
+		statusBar()->message("Sorry, unable to tag the revision");
 }
 
 void MainImpl::ActTagDelete_activated() {
 
 	if (QMessageBox::question(this, "Delete tag - QGit",
-	    "Do you want to un-tag selected revision?",
-	    "&Yes", "&No", QString::null, 0, 1) == 1)
+	                 "Do you want to un-tag selected revision?",
+	                 "&Yes", "&No", QString::null, 0, 1) == 1)
 		return;
 
 	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
 	bool ok = git->deleteTag(lineEditSHA->text());
 	QApplication::restoreOverrideCursor();
-	if (!ok)
-		statusBar()->message("Sorry, unable to un-tag the revision");
-	else
+	if (ok)
 		refreshRepo(true);
+	else
+		statusBar()->message("Sorry, unable to un-tag the revision");
 }
 
 void MainImpl::ActPush_activated() {
@@ -1268,8 +1451,8 @@ void MainImpl::ActPush_activated() {
 	ok = true;
 	QString sha;
 	for (uint i = 0; i < selectedItems.count(); i++) {
-		QString tmp(QString("Pushing patch %1 of %2")
-				.arg(i+1).arg(selectedItems.count()));
+		const QString tmp(QString("Pushing patch %1 of %2")
+		                  .arg(i+1).arg(selectedItems.count()));
 		statusBar()->message(tmp);
 		sha = selectedItems[selectedItems.count() - i - 1];
 		if (!git->stgPush(sha)) {
@@ -1277,12 +1460,12 @@ void MainImpl::ActPush_activated() {
 			break;
 		}
 	}
-	if (!ok)
-		statusBar()->message("Failed to push patch " + sha);
-	else
+	if (ok)
 		statusBar()->clear();
+	else
+		statusBar()->message("Failed to push patch " + sha);
 	QApplication::restoreOverrideCursor();
-	refreshRepo();
+	refreshRepo(false);
 }
 
 void MainImpl::ActPop_activated() {
@@ -1296,7 +1479,7 @@ void MainImpl::ActPop_activated() {
 	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
 	git->stgPop(selectedItems[0]);
 	QApplication::restoreOverrideCursor();
-	refreshRepo();
+	refreshRepo(false);
 }
 
 void MainImpl::ActFilterTree_toggled(bool b) {
@@ -1313,7 +1496,8 @@ void MainImpl::ActFilterTree_toggled(bool b) {
 		rv->treeView->getTreeSelectedItems(selectedItems);
 		if (selectedItems.count() == 0) {
 			if (rv->st.fileName().isEmpty()) {
-				dbs("ASSERT tree filter action activated with no selected items");
+				dbs("ASSERT tree filter action activated "
+				    "with no selected items");
 				return;
 			}
 			// tree could be empty also with exsistant curFileName
@@ -1338,14 +1522,13 @@ void MainImpl::ActFindNext_activated() {
 			return;
 
 		if (endOfDocument) {
-			QMessageBox::warning(this, "Find text - QGit",
-			"Text \"" + textToFind + "\" not found!", QMessageBox::Ok, 0);
+			QMessageBox::warning(this, "Find text - QGit", "Text \"" +
+			             textToFind + "\" not found!", QMessageBox::Ok, 0);
 			return;
 		}
-		if (QMessageBox::question(this, "Find text - QGit",
-			"End of document reached\n\n"
-			"Do you want to continue from beginning?", QMessageBox::Yes,
-			QMessageBox::No | QMessageBox::Escape) == QMessageBox::No)
+		if (QMessageBox::question(this, "Find text - QGit", "End of document "
+		    "reached\n\nDo you want to continue from beginning?", QMessageBox::Yes,
+		    QMessageBox::No | QMessageBox::Escape) == QMessageBox::No)
 			return;
 
 		endOfDocument = true;
@@ -1364,8 +1547,8 @@ void MainImpl::ActFind_activated() {
 		def = te->selectedText().section('\n', 0, 0);
 
 	bool ok;
-	QString str(QInputDialog::getText("Find text - QGit",
-			  "Text to find:", QLineEdit::Normal, def, &ok, this));
+	QString str(QInputDialog::getText("Find text - QGit", "Text to find:",
+	                                  QLineEdit::Normal, def, &ok, this));
 	if (!ok || str.isEmpty())
 		return;
 
@@ -1377,6 +1560,7 @@ void MainImpl::ActHelp_activated() {
 
 	// helpInfo is defined in help.h
 	HelpBase* helpDlg = new HelpBase(NULL, 0, Qt::WDestructiveClose);
+	connect(this, SIGNAL(closeAllWindows()), helpDlg, SLOT(close()));
 	helpDlg->textEditHelp->setText(QString::fromLatin1(helpInfo));
 	helpDlg->show();
 	helpDlg->raise();
@@ -1386,7 +1570,7 @@ void MainImpl::ActAbout_activated() {
 
 	static const char* aboutMsg =
 	"<center><p><b>QGit version " PACKAGE_VERSION "</b></p><br>"
-	"<p>Copyright  2005 Marco Costalba</p>"
+	"<p>Copyright  2005, 2006 Marco Costalba</p>"
 	"<p>Use and redistribute under the "
 	"terms of the GNU General Public License</p></center>";
 	QMessageBox::about(this, "About QGit", QString::fromLatin1(aboutMsg));
@@ -1394,8 +1578,19 @@ void MainImpl::ActAbout_activated() {
 
 void MainImpl::closeEvent(QCloseEvent* ce) {
 
+	// lastWindowClosed() signal is emitted by close(), after sending
+	// closeEvent(), so we need to close _here_ all secondary windows before
+	// the close() method checks for lastWindowClosed flag to avoid missing
+	// the signal and stay in the main loop forever, because this signal is
+	// connected to qApp->quit()
+	//
+	// note that we cannot rely on setting 'this' parent in secondary windows
+	// because when close() is called childern are still alive and, finally,
+	// when childern are deleted, d'tor do not call close() anymore. So we miss
+	// lastWindowClosed() signal in this case.
+	emit closeAllWindows();
+
 	hide();
-	delete commitViewer;
 	delete fv;
 	delete rv;
 
