@@ -18,7 +18,7 @@ class MyProcess : public QProcess {
 Q_OBJECT
 public:
 	MyProcess(QObject *go, Git* g, const QString& wd, bool reportErrors);
-	bool runSync(SCRef runCmd, QString* runOutput, SCRef buf);
+	bool runSync(SCRef runCmd, QString* runOutput, QObject* rcv, SCRef buf);
 	bool runAsync(SCRef rc, QObject* rcv, SCRef buf);
 
 signals:
@@ -28,11 +28,13 @@ signals:
 public slots:
 	void cancel();
 	void myReadFromStdout();
+	void myReadFromStderr();
 	void myProcExited();
 
 private:
 	void setupSignals();
 	bool launchMe(SCRef runCmd, SCRef buf);
+	void sendErrorMsg(bool notStarted = false);
 	void restoreSpaces(QString& newCmd, SCRef sepChar);
 	const QStringList splitArgList(SCRef cmd);
 
@@ -46,6 +48,7 @@ private:
 	bool canceling;
 	bool busy;
 	bool exitStatus;
+	bool async;
 };
 
 #endif
