@@ -7,7 +7,6 @@
 #ifndef SETTINGSIMPL_H
 #define SETTINGSIMPL_H
 
-#include <qsettings.h>
 #include "settingsbase.h"
 
 class Git;
@@ -32,6 +31,7 @@ protected slots:
 	void lineEditTemplate_textChanged(const QString& s);
 	void lineEditCommitExtraOptions_textChanged(const QString& s);
 	void comboBoxCodecs_activated(int i);
+	void pushButtonExtDiff_clicked();
 
 private:
 	void setupCodecList(QStringList& list);
